import UIKit
import Mapbox
import CoreMotion
import AVFoundation

class MainViewController: UIViewController {

    @IBOutlet var map: MGLMapView?
    @IBOutlet var gauge: LevelGauge?

    // We capture motion samples over time to average them.
    var motionSamples = [Double]()

    // Levels run from 0 (unaltered map) to 5 based on Steps.json.
    var currentLevel = 0
    var lastLevel = 0

    // We parse Steps.json into this as well as use it to track active levels.
    var steps = NSMutableDictionary()

    // Motion sample averaging and UI update timer.
    var timer: Timer?

    // Mode toggle sound effect player.
    var clickPlayer: AVPlayer?

    // User motion tracker.
    let motionManager = CMMotionManager()

    // Threshold is for Core Motion activity, Refresh is UI update interval
    let modes = [
        "Walking": [
            "Threshold": 0.1,
            "Refresh": 5
        ],
        "Running": [
            "Threshold": 0.2,
            "Refresh": 5
        ],
        "Testing": [
            "Threshold": 0.05,
            "Refresh": 1
        ]
    ]

    var currentMode = "Walking"

    override func viewDidLoad() {
        super.viewDidLoad()

        map?.styleURL = LushStyleURL
        map?.setCenter(LushStartingCenter, zoomLevel: LushStartingZoom, animated: false)

        map?.tintColor = UIColor.darkGray

        gauge?.labelText = currentMode
        gauge?.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(toggleMode)))

        parseSteps()

        beginMotionTracking()
    }

    func toggleMode(tap: UITapGestureRecognizer) {
        if tap.state == .ended {
            // Play sound effect on devices.
            if TARGET_OS_SIMULATOR == 0 {
                clickPlayer = AVPlayer(url: URL(fileURLWithPath: Bundle.main.path(forResource: "click", ofType: "m4a")!))
                clickPlayer?.play()
            }

            // Suspend motion tracking & reset UI.
            motionManager.stopDeviceMotionUpdates()
            timer?.invalidate()
            currentLevel = 0
            lastLevel = 0
            gauge?.value = 0

            // Reset the map.
            for i in stride(from: 1, through: steps.allKeys.count, by: 1) {
                greenify(index: i, on: false)
            }

            // Toggle the mode.
            switch currentMode {
            case "Walking":
                currentMode = "Running"
            case "Running":
                currentMode = "Testing"
            case "Testing":
                currentMode = "Walking"
            default:
                break
            }
            gauge?.labelText = currentMode

            // Start motion tracking again.
            beginMotionTracking()
        }
    }

    func parseSteps() {
        if let parsedSteps = Utility.mutableDictionaryForJSONResource("Steps", ofType: "json") {
            for step in parsedSteps.allKeys {
                // Add 'greenified' key to each level for current state tracking.
                (parsedSteps[step] as! NSMutableDictionary).setValue(false, forKey: "greenified")
            }
            steps = parsedSteps
        }
    }

    func beginMotionTracking() {
        // Clear samples & track user motion every second.
        motionSamples.removeAll()
        motionManager.deviceMotionUpdateInterval = 1
        motionManager.startDeviceMotionUpdates(to: OperationQueue.main) { [unowned self] (data, error) in
            guard error == nil else {
                self.motionManager.stopDeviceMotionUpdates()
                self.timer?.invalidate()
                let alert = UIAlertController(title: "Motion Query Problem", message: "There was a problem measuring movement. Please force-quit and restart the app.", preferredStyle: .alert)
                alert.addAction(UIAlertAction(title: "OK", style: .default, handler: nil))
                self.present(alert, animated: true, completion: nil)
                return
            }
            // Track user acceleration in the z-plane (in and out of the screen, which assumes the user holds device upright).
            if let value = data?.userAcceleration.z {
                let absoluteValue = abs(value)
                self.motionSamples.append(absoluteValue)
            }
        }

        // Start a new UI refresh timer.
        timer?.invalidate()
        timer = Timer(timeInterval: TimeInterval(modes[currentMode]!["Refresh"] as! Int), repeats: true) { [unowned self] (timer) in
            let stepCount = self.steps.allKeys.count
            // Figure out if user is active or not and increment or decrement the level within bounds.
            if self.average(samples: self.motionSamples) >= self.modes[self.currentMode]!["Threshold"] as! Double {
                self.currentLevel = (self.lastLevel < stepCount ? (self.lastLevel + 1) : self.lastLevel)
            } else {
                self.currentLevel = (self.lastLevel > 0 ? (self.lastLevel - 1) : self.lastLevel)
            }
            // Toss the current samples for this UI refresh.
            self.motionSamples.removeAll()
            // Update the map & the gauge, working up or down the levels.
            if self.currentLevel != self.lastLevel {
                if self.currentLevel > self.lastLevel {
                    for i in stride(from: 1, through: stepCount, by: 1) {
                        if i <= self.currentLevel && !self.stepIsGreenified(index: i) {
                            self.greenify(index: i, on: true)
                        }
                    }
                } else if self.currentLevel < self.lastLevel {
                    for i in stride(from: stepCount, through: 1, by: -1) {
                        if i > self.currentLevel && self.stepIsGreenified(index: i) {
                            self.greenify(index: i, on: false)
                        }
                    }
                }
                self.lastLevel = self.currentLevel
                self.gauge?.value = self.currentLevel
            }
        }
        RunLoop.main.add(timer!, forMode: .commonModes)
    }

    func average(samples: [Double]) -> Double {
        var sum: Double = 0
        samples.forEach { (sample) in
            sum = sum + sample
        }
        return sum / Double(samples.count)
    }

    func stepIsGreenified(index: Int) -> Bool {
        let key = "L\(index)"
        let info = steps[key] as! NSDictionary
        return info.value(forKey: "greenified") as! Bool
    }

    func greenify(index: Int, on: Bool) {
        let key = "L\(index)"
        (steps[key] as! NSMutableDictionary).setValue(on, forKey: "greenified")
        if let step = steps[key] as? NSDictionary,
           let paints  = step["paint"] as? NSArray,
           let layouts = step["layout"] as? NSArray {
            // Retrieve the paint and layout style properties from Steps.json to update for this level.
            [paints, layouts].forEach { (properties) in
                for property in properties {
                    if let property       = property as? NSArray,
                       let layerNames     = property[0] as? [String],
                       let propertyName   = property[1] as? String,
                       let propertyValues = property[2] as? [AnyObject] {
                        for layerName in layerNames {
                            if let layer = map?.style().layer(withIdentifier: layerName) {
                                // Translate e.g. `fill-color` (JS) to `fillColor` (iOS).
                                let propertyKey = Utility.layerPropertyKeyForPropertyName(propertyName: propertyName)
                                // Each property array in the JSON has two values, one drab & one lush.
                                let rawValue = propertyValues[(on ? 1 : 0)]
                                var finalValue: AnyObject?
                                if let testValue = rawValue as? String {
                                    // Parse CSS-style colors if necessary.
                                    if testValue.hasPrefix("hsl") {
                                        finalValue = MGLStyleValue(rawValue: UIColor(hslString: testValue))
                                    } else if testValue.hasPrefix("#") {
                                        finalValue = MGLStyleValue(rawValue: UIColor(hexString: testValue))
                                    }
                                }
                                // Otherwise just wrap raw values (e.g. an image name for `icon-image`).
                                if finalValue == nil {
                                    finalValue = MGLStyleValue(rawValue: rawValue)
                                }
                                // Use KVC on the layer to update the styling properties.
                                let setter = NSSelectorFromString("setValue:forKey:")
                                layer.perform(setter, with: finalValue!, with: propertyKey)
                            }
                        }
                    }
                }
            }
        }
    }

}
